/*
|-- SerialComm.cc    V1.21

Check the | on the left of the comment for new text since previous version.

------------------------------------------------------------------------------
-- Description:

      Visual component class that sets up and manages a serial communication 
      handler for Visual dBase 7.01 applications.

-- Programmer:

|      Nicolas Martin,          nicolas.martin@freesurf.fr

-- Copyright notice:

      This file is owned by Nicolas Martin. See the file "Copyright.txt" for full 
      copyright details and terms of use.

-- History:

      12/07/1998 -- Original V1
      06/10/1998 -- V 1.01 : 
         Clarify comments, add minor improvements, correct minor bugs :
              -- Improvements :
                 + The call to the user ISR when the continuous read is 
                   activated is now performed by a try/catch exception
                   handler, for better robustness.
                 + The fullDuplexWrite function has been simplified and
                   enhanced, based on a different way to process the read
                   operation while sending the data.
                 + flushPort function added, performed by a Win32 API call
                   that completly flushes the com port (read _&_ write).
                 + Added a version property to show within the inspector
                   or the application the current version of the component.
              -- Bugs fixed :
                 + The return code of several functions was incorrect when 
                   the com port was not opened (returned true instead of false)
                 + The continuousRead function has been changed in order to
                   empty the com port read buffer at each tick, to avoid
                   loss of received data due to timer jitter and minimum value.
                   This leads to a slightly different behaviour of that function.
|		  10/10/1999 -- V 1.2 :
|			  Fix several bugs, use structure classes and add new features :
|					-- Bugs fixed : 
|						+ The handshaking method Xon/Xoff and RTS/CTS+DTR/DSR were switched.
|						  In order to keep compatibility with already existing 
|						  applications, the handshaking methods are now changed to :
|							 2 for hardware (default to DTR/DSR+RTS/CTS) handshake
|							 1 for Xon/Xoff handshake
|						+ Trap error when trying to open twice a port without closing
|						  it before.
|                 + Disable parity check if parity = None is selected.
|                 + Sets the Write timeouts to avoid dead lock if handshake is 
|                   selected and the host is not ready.
|					-- New features :
|						+ Full compatibility with V 1.01., previous source code that 
|						  uses serialComm V 1.01 should remain unchanged.
|						  -> Please check the "How to use" section of this comment for
|						  new instructions of use.
|						+ Improved performances in R/W by using win API calls. 
|						+ Improved maintainability by using structure classes that 
|						  enables easy correspondance with the WinAPI C++ structures.
|						+ Surface properties that give informations about the opened 
|						  COM port: 
|						  DOS name, BaudRate, Parity, Data, StopBits, Xon/Xoff config, 
|						  comm port status, comm port errors
|						+ Extend the setHandshake method with new hardware handshake
|						  options.
|						+ Support for communication errors reporting and handling, 
|						  allowing different behaviours of the read and write processes.
|                 + Support for communication status report.
|						+ New method to configure Xon/Xoff handshake.
|                 + Support for communication line TX and RX status.
|		  10/12/1999 -- V 1.21 :
|           Add some controls in the startContinuousRead :
|                 + Check that the number of characters to read is less than
|                   the receive queue size, with a margin of 2x
|                 + Display a warning message in the command window if timer
|                   tick is set to less than 55 ms.

-- designed for Visual dBase, version 7.01, tested under Win 95

|-- dependancies:
|
|			- StructureEx.cc
|			- StructAPI.prg            (_dbWinHome\samples)
|			- structmem.dll				(_dbWinHome\samples)
|			- StructAPI.h             	(_dbWinHome\include)
| 			- winbase.h                (_dbWinHome\include)
|			- winnt.h                  (_dbWinHome\include)
|			- Vdbase.h						(_dbWinHome\include)

-- How to use:
   -----------

      Add the SerialComm.cc component to your Visual dBase component palette
      (See your Visual dBase 7 User Manual if you need help to do that, or consult
      the online help, search for "custom components:adding to palette").
      You can then select the SerialComm component from the Custom page of the
      component palette and simply add it to your form in order to access to the
      component methods. The component appear on the form in design mode, it does
      not appear in run mode, (it behaves somehow like the Query component) so you 
      can deposit it anywhere on your form, even above other objects.
      Or you can choose to create an instance of the SerialComm object by programmation, 
      it would look like :
				
*            set procedure to SerialComm.cc additive
*            this.serialComm1 = new SerialComm( this )

| IMPORTANT BEFORE USE IN VDB'S IDE AND IN A DEPLOYED EXECUTABLE:
| --------------------------------------------------------------
|				1/ Place in the same directory the files:
|					   serialComm.cc
|					   structureEx.cc
|				2/ You need to have the samples directory installed as part of your
|					VdB 7.01 installation.
|				3/ To build an executable that includes serialComm, you need to:
|					- add the following files to your project .prj description file:
|					   serialComm.cc
|					   structureEx.cc
|					   structAPI.prg     (it is in the samples directory; leave it there,
|												  simply reference it in the .prj file)
|					- place a copy of the following .dll in the executable directory
|						(you need to add it to your deployed application files)
|					   structmem.dll	    (it is in the samples directory)
|				4/ If you upgrade from V 1.01, a complete rebuild will be necessary.
						

|- Currently unsupported capabilities, bugs:
|  ----------------------------------------

|  - While in Xon/Xoff handshake, the transmission by Windows seems to ignore 
|    the state of the fTXContinueOnXoff DCB member, assuming it is always true.
|
|  - The ReadFile and WriteFile API functions seem to ignore the state of the 
|		fAbortOnError DCB member, assuming it is always true.
  


--------------------------------------------------------------------------------
                ---   COMPONENT PROPERTIES AND METHODS   ---
--------------------------------------------------------------------------------

- Properties:
  -----------

--  displayMessages
      Flag, indicates whether SerialComm displays internal messages when errors
      occur during execution. displayMessages defaults to true. When setting 
      displayMessages property to false, no error messages will be displayed.
      However, testing the return codes of each Method allows you to know whether
      an error occured during its execution.

|	IMPORTANT NOTE: ALL THE PROPERTIES BELOW ARE READ ONLY AND SHOULD NOT BE
|						 CHANGED DIRECLTY; THEY WILL BE UPDATED BY SERIALCOMM THROUGH 
|						 THE DIFFERENT METHODS AVAILABLE.	

--  version
         String, shows the current version of the serialComm component. This property
         can be seen in the inspector (in the form design mode, for instance) or accessed
         within your program.

|--  ComPort
|			This property is set up after the communication is correctly opened by
|			the openport method. It contains the DOS name for the currently opened 
|			communication port.
|
|-- 	BaudRate
|			This property is set up after the communication is correctly opened by
|			the openport method. It contains the communication speed for the currently
|			opened communication port.
|
|-- 	Parity
|			This property is set up after the communication is correctly opened by
|			the openport method. It contains the parity setting for the currently opened 
|			communication port.
|
|-- 	Data
|			This property is set up after the communication is correctly opened by
|			the openport method. It contains the number of bits per character for the 
|			currently opened communication port.
|
|-- 	StopBits
|			This property is set up after the communication is correctly opened by
|			the openport method. It contains the number of stopbits for the currently opened 
|			communication port.
|
|--	XonXoffSet
|			This property contains information about the current Xon/Xoff configuration.
|			It consists in a VdB object whose properties are :
|					.XonChar
|					.XoffChar
|					.XonPercent					
|					.XoffPercent					
|					.InBufSize	
|
|				.XonChar: a VdB numeric, ASCII code of the currently selected Xon 
|							 char that will be used if Xon/Xoff handshaking is set.
|							 It can be changed using the setXonXoffConfig method (see below).
|				.XoffChar: a VdB numeric, ASCII code of the currently selected Xoff 
|							 char that will be used if Xon/Xoff handshaking is set.
| 							 It can be changed using the setXonXoffConfig method (see below).
|				.XonPercent: VdB numeric, percentage of input buffer size that is  
|								 still filled, before the Xon char is sent.
|				.XoffPercent: VdB numeric, percentage of input buffer size filled that 
|								  will trig the transmission of the Xoff char.
|				.InBufSize: VdB numeric, input buffer size
|                                    
|--	CommErrors
|			The CommErrors property contains information about the communication
|			line errors that have been encountered, and reported by the 
|			checkCommErrors or checkPortStatus methods.
|			This property consists in a VdB object containing itself 2 properties:
|				.type
|				.text
|			These 2 properties are arrays of varying size, which will be updated 
|			by the checkCommErrors or checkPortStatus method.
|        Both will have always the same size.
|			If communication errors have been encountered, a call to 
|			checkCommErrors or checkPortStatus will report these errors in 
|        these 2 arrays as a new entry in each one of them. 
|				.type[i] : a VdB numeric, representing the error type, will be added
|				.text[i] : a VdB string, explaining the error nature, will be added
|			There will be as many different elements added as types of errors encountered. 
|			The reported errors can be:
|			 ----------------------------------------------------------------------
|			Error |			Error        |        			Meaning
|			type  |  		string		 |
|			 ----------------------------------------------------------------------
|				1	| "Break"         	 |	The hardware detected a break condition.
|			 	2	| "Framing            | The hardware detected a framing error.
|				3	| "Char overrun"      | A character-buffer overrun has occurred. 
|					|							 |	The next character is lost.
|				4	| "Parity"            | The hardware detected a parity error.
|				5	| "Transmit overflow" | The application tried to transmit a character, 
|					|							 |	but the output buffer was full.
|				6	| "Receive overflow"  | An input buffer overflow has occurred. 
|					|							 |	There is no more room in the input buffer.
|													
|			Note: If no errors are encountered, the arrays will be empty.
|
|--	PortStatus
|			This property consists in a VdB object containing itself 5 properties:
|				.CTSHold
|				.DSRHold
|				.XoffHold
|				.XoffSent
|				.CharsInRxQ
|				.CharsNotSent
|       These properties are updated each time a call to checkPortStatus is issued.
|			They have the following meanings:
|				.CTSHold : VdB boolean, true means Transmission is waiting for 
|							  the CTS signal
|				.DSRHold : VdB boolean, true means Transmission is waiting for 
|							  the DSR signal
|				.XoffHold : VdB boolean, true means Transmission is waiting because 
|								the XOFF character was received. 
|				.XoffSent : VdB boolean, true means Transmission is waiting because 
|								the XOFF character was transmitted. This is to cope with
|								hosts that takes the next character as XON, regardless 
|								of the actual character.
|				.CharsInRxQ : VdB numeric, the number of chars not yet read in the 
|								  reception queue.
|				.CharsNotSent : VdB numeric, the number of chars not sent when an
|									 error occurs while writing to the comm port. If no
|									 error occurs, the value will be 0.
|			
- Methods:
  --------

--	openPort( <cComPort>, <nBaudRate>, <cParity>, <nData>, <nStopBits> )
      This function opens and initializes the Comm port you want to access.
      By default, the handsaking method is set to "no handshake".
      The parameters are:
          <cComPort>: VdB string, containing the DOS name "COM1", "COM2", "COM3", ...
         <nBaudRate>: VdB Numeric, communication line speed.
|          <cParity>: VdB string, can be "None", "Even", "Odd", "Mark" or simply 
|							'N', 'E', 'O', or 'M'. Now case UNsensitive
             <nData>: VdB numeric, number of bits per characters
         <nStopBits>: VdB numeric, can be : 1 or 1.5, or 2
|				Remarks: the following restrictions apply to the values specified 
|							for the data and stopBits : 
|								The number of data bits must be 5 to 8 bits. 
|								The use of 5 data bits with 2 stop bits is an invalid 
|                       combination, as is 6, 7, or 8 data bits with 1.5 stop bits. 
      The function returns true if the port has been opened, and false otherwise

--	closePort( )
      This function closes a previously opened communication port. No parameter 
      is required, the Comm port that has been opened by the current object will
      be closed.
      The function returns true if the port has been closed, or false otherwise

--	setHandshake( <nHSmethod> )
      Sets the Handshaking method for the communication flow control, for read
      and write operations.
          <nHSmethod>: VdB numeric, handshaking method, can be:
								 0 for no handshake, 
|								 1 for Xon/Xoff handshake.
|								 2 for hardware (DTR/DSR + RTS/CTS) handshake.
|   		      			 3 for hardware (DTR/DSR only) handshake.
|   		       			 4 for hardware (RTS/CTS only) handshake.
      The function returns true if it has been correctly executed, or false otherwise

|--	setXonXoffConfig( <nXonChar>, <nXoffChar>, [<nXonPct>], [<nXoffPct>], [<nInBuffSize>] )
|			This method will set up some parameters concerning the Xon/Xoff 
|			handshake configuration. These parameters need to be set before
|			entering the Xon/Xoff handshake mode by a call to setHandshake(1).
|				<nXonChar> : VdB numeric, the new Xon's ASCII code to use. e.g. 0x11
|				<nXoffChar>: VdB numeric, the new Xoff's ASCII code. e.g. 0x13
|				<nXonPct>: Optional parameter : VdB numeric,  default = 20 when port 
|                      is opened.
|                      If omitted, the last value is kept.
|							  If specified, represents the percentage limit of input 
|                      buffer not yet read before the Xon char is sent 
|                      (if an Xoff has been previously sent).
|				<nXoffPct>: Optional parameter : VdB numeric,  default = 80 when port 
|                        is opened.
|                        if omitted, the last value is kept.
|								 If specified, represents the percentage limit of input 
|                        buffer not yet read before the Xoff char is sent.
|				<nInBuffSize>: Optional parameter : VdB numeric,  defaults to Windows default
|									settings.
|									The requested input buffer size, in bytes. Windows may
|									however use a different size than requested.

|				  ex : Default setting for nXonPct and nXoffPct
|             ----
|											Input (RX) buffer
|                   Empty                                Full
|   Data from line -> || -> to application
|                            |                      |
|                      nXonPct = 20            nXoffPct = 80
|
|      The function returns true if it has been correctly executed, or false otherwise

|--  checkCommErrors
|			This method checks the port for the last communication errors (if any).
|			It updates the .CommErrors property if new errors appear, otherwise the 
|			CommErrors object is left unchanged (it is not not cleared).
|			As opposed to checkPortStatus(), checkCommErrors() performs very little
|			computation when there has been no errors, and can thus be called after 
|        each data read.
|			This function returns true if no new errors are encountered, false otherwise.
|
|--	checkPortStatus
|			Checks the communication port to get the last communication errors
|			(if any) and port status.
|        Unlike checkCommErrors, this function will update the following set of 
|			properties of the component, even if no new errors are encountered:
|				.CommErrors
|				.PortStatus
|			For an explanation of each one of them, please refer to the --properties
|			section of the comment.
|      	The function returns true if no communication errors are currently reported,
|		  	(# of Comm Errors = 0) and false if there are error(s).

--	readString( <nMaxChars> )
      One shot read.
      Reads the data that has already been received on the communication port.
      If no data is available, the function returns with an empty string.
      If data is present, it is placed in the returned string.
          <nMaxChars> : VdB numeric, maximum number of characters to read. The 
*                       function will read a number of characters smaller or equal 
                        to <MaxChars>.
      The function will return a string with the data read, and with length equal 
      to the numbers of characters read, or an empty string.

--	startContinuousRead( <nBufferSize>, <fpCallBack> )
      This high level function has been designed in order to perform Full Duplex
      communications with very little programmatic effort.
      The goal is to set the SerialComm component in reception mode so that 
      it behaves somehow like an Interrupt handler, notifying a user Interrupt 
      Service Routine (ISR) as soon as new characters have been received on the 
      reception line. Actually, after the startContinuousRead function has been 
      executed, the following activity is being initialized :
      - A VdB timer is set up, and fires regularly to look whether new 
        characters have been received on the reception line.
      - If new characters are available, the new characters are placed in 
        a string and passed as a parameter to the user ISR (callBack routine)
        which is then called. The way it performs is closer to a Multi-Thread 
        application than to an Interrupt Handler, but the result is roughly the same, 
        with some particularities explained below.
        The parameters passed to startContinuousRead are :
          <nBufferSize> : VdB numeric, specifies the desired maximum number of characters
                         that should be passed in the parameter string to the ISR.
                         In fact, more or less characters can be passed each time to the 
                         ISR, depending on the count actually received, and on the timer
                         jitter.
          <fpCallBack> : VdB function pointer, of the ISR that is in charge 
                       of processing the characters read from the port.
                       This ISR must be declared with a parameter, which is the 
                       name of the string that contains the new characters received.
                       For instance, if the ISR is part of a form, it can be 
                       declared as :

*                        function processReadData( newChars )
                           ...
                           process the characters in newChars
                           ...
                           return

                       to initialize the read thread, you can place a :

                           this.serialComm1.startContinuousRead( 50, this.processReadData )

                       in the form's onOpen method, to get typically 50 chars in newChars.

         The function returns true if it has been correctly initialized, or false otherwise

      Important Remarks and Recommandations
      -------------------------------------
      	1/ Serial communication line arithmetics
      	As described above, the read thread uses a timer to check for the presence
      	of new characters. The timer tick is set to a value corresponding to the
      	minimum time to get a number of characters equal to <bufferSize>, given 
      	the parameters used to open the port. This value (in seconds) is :

      	<nBufferSize> * ( 1 + <nData> + <0 or 1 parity bit> + <nStopBits> ) / <nBaudRate>

      	Note: a character always contains 1 start bit, additionnaly to the 
      	data, the parity and stop bits.
					
      	In order to get efficient software, you need to set a <bufferSize> not
      	too small when <BaudRate> gets high. Furthermore, your user-defined ISR
      	execution time must be smaller than the timer value.
      	For instance, with :
      	<bufferSize> = 150, <data> = 8, <stopBits> = 1, <BaudRate> = 19200, <Parity> = 'None'
      	the timer tick will be set to 78 milliseconds.

      	IMPORTANT : the minimum timer tick handled by VdB is 55 ms. If the timer
      	---------   tick calculated above in the startContinuousRead function is 
                     lower, the minimum value of 55 ms will used by VdB. In that
                     case, a Warning message is now issued in the command window.
                  	In all the cases, the port will be read as much as necessary to 
                  	empty it. This may lead to having always more characters
                  	than <bufferSize> at each ISR call.

      	2/ Full duplex recommandation
      	The writeString operation is performed in one block, this means that the 
      	writeString function will not be interrupted until all the chars have
      	been sent on the port.
      	Thus, when performing Full Duplex communications, you should not send
      	a long string in one block with writeString, but rather use the function
      	fullDuplexWrite, which enables read and write at the same time.

					
--	stopContinuousRead( )
      	After a startContinuousRead, the acquisition goes on until either :
      	- a stopContinuousRead is performed, 
      	- or the Comm port is closed.
      	The function returns true if it has been correctly executed, or false otherwise

--	writeString( <cWstring> )      
|      	Sends the data in <Wstring> on the communication line in half duplex mode.
         The function will send all the data contained in Wstring.
          <cWstring>: VdB string, the buffer of characters to send.
|		   If an error occurs, generally when a protocol is set (Xon/Xoff or 
|			hardware), and the host requests to stop transmission (Xoff or hardware 
|        signal) the function will return false, and update the .charsNotSent 
|			property of the PortStatus object with the number of chars from <cWstring>
|			that were not sent.
|      	The function returns true after all the characters have been sent, and 
|        false if not all characters have been sent. In that case, the 
|        .PortStatus.CharsNotSent property will give the # of chars not sent.

--	fullDuplexWrite( <cWstring> )
|      This function sends the data in <Wstring> on the communication line, in
|      Full Duplex mode.
|      In that case, characters of <cWString> are sent using writeString, while
|      incoming ones are alternatively read using 'readString', if continuous read 
|      was  previously enabled. 
       When characters are available in reception, the user ISR is called 
       the same way as defined for 'startContinuousRead'.
            <cWstring> : VdB string, contains the characters to be sent.
|      fullDuplexWrite performs calls to writeString, so the same remark applies 
|      when a transmission error occurs.
       The function returns true if all the characters have been sent, and 
       false otherwise.
			

--	sendBreak( <duration> )
      	This enables the user to send a communication BREAK on the serial line.
      	A communication BREAK is typically sent in order to stop the communication
      	process by either one of the two devices that are at each end of the line.
      	The duration of the BREAK command can be specified.
        <duration> : VdB numeric, specifies the duration (in seconds) of the BREAK.
                     If no parameter is passed to this function, a default BREAK
                     of 1 second is sent on the line.
      	The function returns true if it has been correctly executed, or false otherwise.

--	flushPort( )
      	This function flushes the read and write buffers from any characters.
      	The function returns true if it has been correctly executed, or false otherwise.

--	flushReadPort( )
      	This function flushes the reception line from any characters.
         You should rather use the flushPort method instead, this function has
         been left just for backward compatibility.
      	The function returns true if it has been correctly executed, or false otherwise.

------------------------------------------------------------------------------
*/
// Here starts the code :

// include VdB files
#include "winbase.h"
#include "winnt.h"
#include "Vdbase.h"

class SerialComm( f ) of Rectangle( f ) custom 
// the following protections makes properties and methods hidden to the user.
		protect ;
		printable, visible, text, fontSize, fontStrikeOut, ;
		border, borderStyle, colorNormal, mousePointer, pageNo, patternStyle, ;
		onDesignOpen, onLeftDblClick, onLeftMouseDown, onLeftMouseUp, ;
		onMiddleDblClick, onMiddleMouseDown, onMiddleMouseUp, onMouseMove, onOpen, ;
		onRightDblClick, onRightMouseDown, onRightMouseUp, ;
		chk, serialOnClose, serialCommOnOpen, timerLoop, nativeSessionTimerLoop, lpMess, ;
		IsPortOpen, nBufferSize, cByteBuffer, fpCallBack, SCtimer, secondPerChar, ;
		SCpaintBox, ComStat, CommProp , dcb, CommTimeouts
		
// define constants
#define FILE_NOT_SHARED		0
#define MINUS_ONE				0xffffffff
#define BIGBUFFERSIZE		10000
#define ERRORTITLE			'Error in SerialComm ...'
#define DEFAULT_XON 			0x11
#define DEFAULT_XOFF			0x13

	// configure this object

	with ( this )
		onOpen := this.SerialCommOnOpen
		text := 'SerialComm1.21'
		height := 1
		width := 15
		visible := false
		colorNormal := 'yellow/navy'
	endwith

	// set the custom properties
	this.displayMessages = true						// default set to true, display internal messages 
	this.version = 'SerialComm V 1.21'				// version identification
	this.ComPort = Null          						// DOS name for the com port
	this.BaudRate = Null									// BaudRate
	this.Parity = Null									// Parity, if any
	this.Data = Null										// Number of data bits
	this.StopBits = Null									// Number of stop bits
	this.XonXoffSet = new XonXoffConfigClass( )	// contains info about the Xon/Xoff configuration
	this.CommErrors = new CommErrorClass( )		// contains info about the last errors detected
	this.PortStatus = new PortStatusClass( )		// contains info about the port status

	// reference to ComPort : initialized to no port opened
	this.hSerial = MINUS_ONE			

	// include declarative structure classes for the StructureEx.cc class
	private sFolder
	sFolder = program( 0 )
	sFolder = left( sFolder, sFolder.lastIndexOf( '\' ))
	set procedure to "&sFolder.\StructureEx.cc" additive

   this.ComStat = new _COMSTAT()				   // Creates a ComStat structure
	this.dcb = new _DCB()                		// Creates a DCB structure
	this.CommProp = new _COMMPROP()          	// creates a COMMPROP structure

	// Create the timer for the continuous read function
	this.SCtimer = new timer( )

	// loads the API 32 functions used by SerialComm
	extern HANDLE CreateFile(LPCSTR, DWORD, DWORD, LPSTRUCTURE, DWORD, ;
						DWORD, HANDLE) kernel32 from "CreateFileA"
	extern BOOL   CloseHandle(HANDLE) kernel32
	extern BOOL   GetCommState(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   SetCommState(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   BuildCommDCB(LPCSTR, LPSTRUCTURE) kernel32 from "BuildCommDCBA"
	extern BOOL   WriteFile(HANDLE, LPCVOID, DWORD, LPDWORD,LPSTRUCTURE) kernel32
	extern BOOL   ReadFile(HANDLE, LPVOID, DWORD, LPDWORD,LPSTRUCTURE) kernel32
	extern BOOL   GetCommTimeouts(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   SetCommTimeouts(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   SetCommBreak(HANDLE) kernel32
	extern BOOL   ClearCommBreak(HANDLE) kernel32
	extern BOOL   PurgeComm(HANDLE, DWORD) kernel32
	extern BOOL   ClearCommError(HANDLE, LPDWORD, LPSTRUCTURE) kernel32
	extern BOOL   GetCommProperties(HANDLE, LPSTRUCTURE) kernel32
   extern BOOL   SetupComm(HANDLE, DWORD, DWORD) kernel32

   // Loads API 32 functions for MultiByte to WideChar conversions. These ones 
   // are not prototyped in _DBWINHOME\Include\win32API.prg
   extern CINT	  MultiByteToWideChar( UINT, DWORD, CPTR, CINT, CPTR, CINT ) kernel32
	extern CINT   WideCharToMultiByte( UINT, DWORD, CPTR, CINT, CPTR, CINT, LPCSTR, LPBOOL) kernel32

	//-------------------------- serialCommOnOpen -------------------------------
	// Linked method : Serial Comm On open event
	function serialCommOnOpen
		// the paintbox created here tracks an OnClose event, for checking whether
		// the Com port has been properly closed before closing Serial Comm.
		// These lines can be safely commented to desactivate this capability.
		this.SCPaintBox = new paintbox( this.parent )
		with ( this.SCPaintBox )
			transparent := true
			visible := false
			onClose = this.serialOnCLose
		endwith
		this.SCPaintBox.SerialCommRef = this		// keep track of serialComm object ref
		return

	//------------------------------ openPort -----------------------------------
   // this function opens the communication port
	function openPort( cComPort, nBaudRate, cParity, nData, nStopBits )
		local CommTimeouts, cParityLetter, cParityProper, cComPortUpper, nWriteTimeout

      // Test for a correct combination of data and stop bits
      // ...'The use of 5 data bits with 2 stop bits is an invalid combination, 
      //     as is 6, 7, or 8 data bits with 1.5 stop bits.'...
      if (( nData == 5 ) and ( nStopBits == 2 )) or ;
         (( nData > 5 ) and ( nStopBits == 1.5 ))
			if this.displayMessages
				AlertMessage( 'Invalid combination of data and stop bits.', ERRORTITLE )
			endif
         return false
      endif

      // get parity initial letter, and com port uppercase
		cParityLetter = cParity.left( 1 ).toUpperCase( )
      cParityProper = cParity.toProperCase( )
      cComPortUpper = cComPort.toUpperCase( )

		// Compute the acquisition time per character, for continuous read timer tick
		this.secondPerChar = ( 1 + nData + iif( cParityLetter == 'N', 0, 1 ) ;
										 + nStopBits ) / nBaudRate

		// Compute the write timeout in milliseconds: 2 chars, min = 1 millisecond
		nWriteTimeout = 1 + this.secondPerChar * 2000

		// To open the port, we need to create the file handle to the communication port
		if ( this.hSerial == MINUS_ONE )						// No port opened yet
			this.hSerial = CreateFile( cComPort ,;
									GENERIC_READ + GENERIC_WRITE ,;
									FILE_NOT_SHARED ,;
									Null ,;
									OPEN_EXISTING ,;
									FILE_ATTRIBUTE_NORMAL ,;
									Null )							// opens communication port
		else
			if this.displayMessages
				AlertMessage( 'Cannot open port. ' + this.name + ' Already hooked to a COM port!', ERRORTITLE )
			endif
			return false		// Abort OpenPort
		endif

		if ( this.hSerial == MINUS_ONE )	// Checks it is opened correctly
			if this.displayMessages
				AlertMessage( 'Unable to open COM port', ERRORTITLE )
			endif
			return false		// Abort OpenPort
		endif

		// Now, we set the communication port parameters

		// Get the current DCB values
		if not this.chk( GetCommState( this.hSerial, this.dcb.value ))  
			return false
		endif

		// Here, the old method using BuildCommDCB has been abandonned, the
		// port parameters are now directly set into the DCB structure.
		this.dcb.setMember( "BaudRate", nBaudRate )
		this.dcb.setMember( "fParity", 1 )        				// Enables parity checking
		do case
			case cParityLetter = 'N'
				this.dcb.setMember( "Parity", NOPARITY	)
				this.dcb.setMember( "fParity", 0 )        		// Disables parity check if "None"
			case cParityLetter = 'O'
				this.dcb.setMember( "Parity", ODDPARITY )
			case cParityLetter = 'E'
				this.dcb.setMember( "Parity", EVENPARITY )
			case cParityLetter = 'M'
				this.dcb.setMember( "Parity", MARKPARITY )
		endcase
		this.dcb.setMember( "ByteSize", nData )
		do case
			case nStopBits = 1
				this.dcb.setMember( "StopBits", ONESTOPBIT )
			case nStopBits = 1.5
				this.dcb.setMember( "StopBits", ONE5STOPBITS )
			case nStopBits = 2
				this.dcb.setMember( "StopBits", TWOSTOPBITS )
		endcase

		// Sets the DCB values
		if not this.chk( SetCommState( this.hSerial, this.dcb.value ))  
			return false
		endif

		// Here, we set a default NO Handshake for the Handshaking method.
		this.setHandshake( 0 )

		// Last, we build the timeouts for read and write operations.
		// ..."A value of MAXDWORD for ReadIntervalTimeout, combined with zero 
		// values for both the ReadTotalTimeoutConstant and ReadTotalTimeoutMultiplier 
		// members, specifies that the read operation is to return immediately with 
		// the characters that have already been received, even if no characters 
		// have been received"...
		CommTimeouts = new _COMMTIMEOUTS()          // creates a COMMTIMEOUTS structure

		with( CommTimeouts )
			setMember( "ReadIntervalTimeout", MAXDWORD )
			setMember( "ReadTotalTimeoutMultiplier", 0 )
			setMember( "ReadTotalTimeoutConstant", 0 )

		// Set up write timeouts to avoid dead lock if host device is not ready
		// and handshake is used. Delay is set to 2 chars
			setMember( "WriteTotalTimeoutMultiplier", nWriteTimeout )
			setMember( "WriteTotalTimeoutConstant", nWriteTimeout )
		endwith

		// Sets CommTimeouts
		if not this.chk( SetCommTimeouts( this.hSerial, CommTimeouts.value )) 
			return false
		endif

		// Update the object properties with the port configuration parameters.
		with ( this )
			ComPort = cComPortUpper
			BaudRate = nBaudRate
			Parity = cParityProper
			Data = nData
			StopBits = nStopBits
		endwith

		if not this.chk( getCommProperties( this.hSerial, this.CommProp.value ))
			return false
		endif

		// Sets the Xon/Xoff default configuration
		this.setXonXoffConfig( DEFAULT_XON, DEFAULT_XOFF, 20, 80, ;
									  this.CommProp.getMember( "dwCurrentRxQueue" ))

		// Get and update the comm port status
		this.checkPortStatus( )

		// for debug, see what do DCB, CommTimeouts CommProp look like
		//this.CommTimeouts.dumpMembers()
		//this.dcb.dumpMembers()
		//this.CommProp.dumpMembers()
		return true

	//-------------------------------- closePort --------------------------------
	// this function closes the previously opened communication port
	function closePort
		if this.IsPortOpen( )
			this.stopContinuousRead( )
			if not this.chk( CloseHandle( this.hSerial ))
				return false
			endif
			this.hSerial = MINUS_ONE
		endif
		return true

	//------------------------------ setHandshake -------------------------------
	// this function sets the HandShaking method
	function setHandshake( HSmethod )
		if not this.IsPortOpen( )
			return false
		endif

		do case
			case ( HSmethod == 0 )		// No handshake method
				// ..."fInX, fOutX,fOutXDsrFlow,and fOutXCtsFlow are all set to FALSE
				//fDtrControl is set to DTR_CONTROL_ENABLE
				//fRtsControl is set to RTS_CONTROL_ENABLE"...
				with ( this.dcb )
					setMember( "fInX", 0 )
					setMember( "fOutX", 0 )
					setMember( "fOutxDsrFlow", 0 )
					setMember( "fOutxCtsFlow", 0 )
					setMember( "fDtrControl", DTR_CONTROL_ENABLE )
					setMember( "fRtsControl", RTS_CONTROL_ENABLE )
				endwith

			case ( HSmethod == 1 )		// CHANGED (see comment) to : Xon/Xoff handshake
				// ..."fInX, fOutX are both set to TRUE
				//fOutXDsrFlow,fOutXCtsFlow are both set to FALSE.
				//fDtrControl is set to DTR_CONTROL_ENABLE
				//fRtsControl is set to RTS_CONTROL_ENABLE"...
				with ( this.dcb )
					setMember( "fInX", 1 )
					setMember( "fOutX", 1 )
					setMember( "fOutxDsrFlow", 0 )
					setMember( "fOutxCtsFlow", 0 )
					setMember( "fDtrControl", DTR_CONTROL_ENABLE )
					setMember( "fRtsControl", RTS_CONTROL_ENABLE )
				endwith

			case ( HSmethod == 2 )		// CHANGED (see comment) to : Hardware handshake RTS/CTS+DTR/DSR
				// ..."fInX, fOutX are both set to FALSE
				//fOutXDsrFlow,fOutXCtsFlow are both set to TRUE.
				//fDtrControl is set to DTR_CONTROL_HANDSHAKE
				//fRtsControl is set to RTS_CONTROL_HANDSHAKE"...
				with ( this.dcb )
					setMember( "fInX", 0 )
					setMember( "fOutX", 0 )
					setMember( "fOutxDsrFlow", 1 )
					setMember( "fOutxCtsFlow", 1 )
					setMember( "fDtrControl", DTR_CONTROL_HANDSHAKE )
					setMember( "fRtsControl", RTS_CONTROL_HANDSHAKE )
				endwith

			case ( HSmethod == 3 )		// NEW (see comment) : Hardware handshake DTR/DSR only
				// ..."fInX, fOutX are both set to FALSE
				//fOutXDsrFlow is TRUE, fOutXCtsFlow is FALSE.
				//fDtrControl is set to DTR_CONTROL_HANDSHAKE
				//fRtsControl is set to RTS_CONTROL_ENABLE"...
				with ( this.dcb )
					setMember( "fInX", 0 )
					setMember( "fOutX", 0 )
					setMember( "fOutxDsrFlow", 1 )
					setMember( "fOutxCtsFlow", 0 )
					setMember( "fDtrControl", DTR_CONTROL_HANDSHAKE )
					setMember( "fRtsControl", RTS_CONTROL_ENABLE )
				endwith

			case ( HSmethod == 4 )		// NEW (see comment) : Hardware handshake RTS/CTS only
				// ..."fInX, fOutX are both set to FALSE
				//fOutXDsrFlow is FALSE, fOutXCtsFlow is TRUE.
				//fDtrControl is set to DTR_CONTROL_ENABLE
				//fRtsControl is set to RTS_CONTROL_HANDSHAKE"...
				with ( this.dcb )
					setMember( "fInX", 0 )
					setMember( "fOutX", 0 )
					setMember( "fOutxDsrFlow", 0 )
					setMember( "fOutxCtsFlow", 1 )
					setMember( "fDtrControl", DTR_CONTROL_ENABLE )
					setMember( "fRtsControl", RTS_CONTROL_HANDSHAKE )
				endwith

			// If needed, one can add here additional handshaking methods; the dcb 
			// changes can be set the same way as for the 5 methods proposed above.

			otherwise
				if this.displayMessages
					AlertMessage( 'Bad handshake method requested. can be 0-4 only.', ERRORTITLE )
				endif
				return false

		endcase
		if not this.chk( SetCommState( this.hSerial, this.dcb.value ))  // Sets DCB values
			return false
		endif

      // For debug
		// this.dcb.dumpMembers( )
		return true

	//-------------------------- setXonXoffConfig -------------------------------
	// This function sets the Xon/Xoff chars, and optinnaly the Xon and Xoff limits 
   // (as a percentage of the input buffer size) that trig their transmission, 
   // and the size of the receive buffer in bytes.
	function setXonXoffConfig( nXonChar, nXoffChar, nXonPct, nXoffPct, nInBuffSize )
   	local nArgCount

      nArgCount =  ArgCount( )

		if not this.IsPortOpen( )
			return false
		endif		

		if nArgCount >= 5
			if not this.chk( SetupComm( this.hSerial, nInBuffSize, this.CommProp.getMember( "dwCurrentTxQueue" )))
				return false
			endif
			if not this.chk( getCommProperties( this.hSerial, this.CommProp.value ))
				return false
			endif
			this.XonXoffSet.InBufSize = nInBuffSize
		endif

		with ( this.dcb )
			setMember( "XonChar", nXonChar)
			setMember( "XoffChar", nXoffChar )
			setMember( "fTXContinueOnXoff", 0 ) //iif( fTxAfterXoffSent, 1, 0 ))
			if nArgCount >= 3
				setMember( "XonLim", nXonPct * this.XonXoffSet.InBufSize / 100 )
			endif
			if nArgCount >= 4
				setMember( "XoffLim", ( 1 - nXoffPct / 100 ) * this.XonXoffSet.InBufSize )
			endif
		endwith

		if not this.chk( SetCommState( this.hSerial, this.dcb.value ))  	// Sets the new DCB values
			return false
		endif

		with ( this.XonXoffSet )
			XonChar = nXonChar
			XoffChar = nXoffChar
			//TxAfterXoffSent = fTxAfterXoffSent

			if nArgCount >= 3
				XonPercent = nXonPct
			endif

         if nArgCount >= 4
      	   XoffPercent = nXoffPct 
         endif

		endwith

		return true

	//------------------------------ checkCommErrors ----------------------------
	// Checks the comm errors that occured since last call and updates the CommErrors
	// object if new ones have appeared.
	function checkCommErrors
		local nErrorByte

		if not this.IsPortOpen( )
			return false
		endif

		nErrorByte = 0
		if not this.chk( ClearCommError( this.hSerial, nErrorByte, this.ComStat.value ))
			return false
		endif

		if nErrorByte # 0
			this.CommErrors.type = new array( )            // Clear arrays
			this.CommErrors.text = new array( )
			if ( bitand( nErrorByte, CE_BREAK ) # 0 )
				this.CommErrors.type.add( 1 )
				this.CommErrors.text.add( "Break" )
				return false                   // Break : do not check comm errors
			endif
			if ( bitand( nErrorByte, CE_FRAME ) # 0 )
				this.CommErrors.type.add( 2 )
				this.CommErrors.text.add( "Framing" )
			endif
			if ( bitand( nErrorByte, CE_OVERRUN ) # 0 )
				this.CommErrors.type.add( 3 )
				this.CommErrors.text.add( "Char overrun" )
			endif
			if ( bitand( nErrorByte, CE_RXPARITY ) # 0 )
				this.CommErrors.type.add( 4 )
				this.CommErrors.text.add( "Parity" )
			endif
			if ( bitand( nErrorByte, CE_TXFULL ) # 0 )
				this.CommErrors.type.add( 5 )
				this.CommErrors.text.add( "Transmit overflow" )
			endif
			if ( bitand( nErrorByte, CE_RXOVER ) # 0 )
				this.CommErrors.type.add( 6 )
				this.CommErrors.text.add( "Receive overflow" )
			endif

			return false                      // Error
		endif
		return true          // no errors

	//--------------------------- checkPortStatus -------------------------------
	// This method updates the CommErrors and PortStatus properties
	// of the serialComm component.
	function checkPortStatus
		if not this.IsPortOpen( )
			return false
		endif

		if this.CheckCommErrors( )                        // No errors occured
  			this.CommErrors.type = new array( )            // so clear arrays
			this.CommErrors.text = new array( )
      endif

		with ( this.PortStatus )
			CTSHold := ( this.ComStat.getMember( "fCtsHold" ) == 1 )
			DSRHold := ( this.ComStat.getMember( "fDsrHold" ) == 1 )
			XoffHold := ( this.ComStat.getMember( "fXoffHold" ) == 1 )
			XoffSent := ( this.ComStat.getMember( "fXoffSent" ) == 1 ) 
			CharsInRxQ := this.ComStat.getMember( "cbInQue" )
      endwith

      // debug
      //this.ComStat.dumpMembers()
		return true

	//-------------------------------- readString -------------------------------
	// This function reads data from the communication port.
	// The trick here is that we need to convert the byte string returned by 
   // ReadFile, into a VdB Unicode string which is stored as a 16 bits word 
   // string, one character in each word. 
	// This operation is now speed up by using the MultiByteToWideChar API function
	function readString( nMaxChars )
		local cBuffer, cWbuffer, nBytesRead
		nBytesRead = 0		// will be set by ReadFile function
		if this.IsPortOpen( )
			cBuffer = replicate( chr( 0 ), nMaxChars )	// creates the byte buffer
			if not this.chk( ReadFile( this.hSerial, cBuffer, nMaxChars, nBytesRead, Null ))
				return ''
			endif
			// Now, convert the byte string into a VdB standard Unicode string
			cWbuffer = replicate( chr( 0 ), nBytesRead )		
			MultiByteToWideChar( 0, 0, cBuffer, nBytesRead, cWbuffer, nBytesRead )

			return cWbuffer
		endif
		return ''

	//--------------------------- startContinuousRead ---------------------------
	// this function starts the continuous read operation.
	function startContinuousRead( nBufferSize, fpCallBack )
		if not this.IsPortOpen( )
			return false
		endif

      // Check the receive queue size versus the number of characters to read
      // each time, with a margin of 2x
      if ( this.XonXoffSet.InBufSize < 2 * nBufferSize )
			if this.displayMessages
				AlertMessage( 'Receive queue size is too small.', ERRORTITLE )
			endif
         return false
      endif

      // Check the timer interval, so that it is less than VdB's lower limit of 55 ms
      if ( this.secondPerChar * nBufferSize < 0.055 )
         // Tolerated, but display a warning in the command window
			? 'Warning from SerialComm : Continuous read interval has been requested' + ;
           ' to be less than 55 ms. It will use 55 ms.'
      endif

		// creates the buffer for continuous read operations
		this.cByteBuffer = replicate( chr( 0 ), nBufferSize )	// byte buffer
		this.nBufferSize = nBufferSize  			// typical # of chars
		this.n2xBufferSize = 2 * nBufferSize  	// # of chars allowed to receive
		this.fpCallBack = fpCallBack		// saves callback routine name
		this.SCtimer.interval := this.secondPerChar * nBufferSize
		this.SCtimer.NSTL = this.nativeSessionTimerLoop
		this.SCtimer.onTimer := this.timerLoop
		this.SCtimer.enabled := true		// starts the read thread.

		return true

	//---------------------------- timerLoop ------------------------------------
	// this function is used as the continuous read timer loop
	function timerLoop
		return this.NSTL( )

	//---------------------------- nativeSessionTimerLoop -----------------------
	// this function has been created in order to execute the timer loop 
	// steps inside the original form's session. This has been added to
	// keep the form's session settings during loop execution, and especially
	// for the SET TALK primitive
	function nativeSessionTimerLoop
		local cWbuffer, cWtempBuffer, nBytesRead

		nBytesRead = 0		// will be set by ReadFile function
		cWbuffer = ''

		do    // read as much as needed to empty read port buffer
			if not this.chk( ReadFile( this.hSerial, this.cByteBuffer, this.n2xBufferSize, nBytesRead, Null ))
				return false
			endif
			// Now, convert the byte string into a VdB standard Unicode string
			cWtempBuffer = replicate( chr( 0 ), nBytesRead )		
  			MultiByteToWideChar( 0, 0, this.cByteBuffer, nBytesRead, cWtempBuffer, nBytesRead )

			cWbuffer += cWtempBuffer
		until ( nBytesRead < this.n2xBufferSize )

		if cWbuffer.length > 0	// New characters read ?
			try	// try to call the user ISR
				this.fpCallBack( cWbuffer )		
			catch ( Exception e1 )
				this.stopContinuousRead( )	// aborts continuous read
				if this.displayMessages
					AlertMessage( 'Error while calling or executing ISR function', ERRORTITLE )
				endif
				return false
			endtry
		endif
		return true

	//-------------------------- stopContinuousRead -----------------------------
	// this function stops the continuous read operation.
	function stopContinuousRead
		if not this.IsPortOpen( )
			return false
		endif
		this.SCtimer.enabled := false
		return true

	//------------------------------ writeString --------------------------------
	// This function writes a VdB string to the communication port.
	// The same trick applies here as for the read operation, convertion of the VdB 
	// Unicode string to send, stored as a 16 bits word string with one character in each word, 
	// into a byte string containing the characters to send.
	// This operation is now speed up by using the WideCharToMultiByte API function
	function writeString( cWstring )
		local nBytesWritten, cByteStr, nLength
		if not this.IsPortOpen( )
			return false
		endif
		nBytesWritten = 0		// will be set by WriteFile
		nLength = cWstring.length

		cByteStr = replicate( chr( 0 ), nLength )  // creates byte string
      WideCharToMultiByte(0, 0, cWstring, nLength, cByteStr, nLength, Null, Null)

		if not WriteFile( this.hSerial, cByteStr, nLength, nBytesWritten, Null )
			this.PortStatus.CharsNotSent = nLength - nBytesWritten
			return false
		else
			this.PortStatus.CharsNotSent := 0      // All chars have been sent
		endif
		return true


	//-------------------------- fullDuplexWrite --------------------------------
	// This function it used to send data in Full Duplex operation.
	// The string Wstring to be sent is splited and sent in groups of characters, each
	// group containing the number of chars specified at the startContinuousRead.
	// Between each group transmission, the reception line is checked (if continuous
	// read has previously been set ) and user ISR called if new chars are arrived.
	function fullDuplexWrite( cWstring )
		local j, fContReadStart, nBufferSize, cSubString, nCharsSent
		if not this.IsPortOpen( )
			return false
		endif

		fContReadStart = this.SCtimer.enabled			// records that continuous read is started
		nBufferSize = iif( fContReadStart, this.nBufferSize, cWstring.length	)	// if not read continuous started, send string in 1 block
		this.SCtimer.enabled := false		// stops timer ticks and performs polling read

		// processes the string, sending each time nBufferSize characters
      nCharsSent = 0
		for j = 0 to cWstring.length - 1 step nBufferSize
         cSubString = cWstring.subString( j, j + nBufferSize )

			if not this.writeString( cSubString )   // Error
				this.SCtimer.enabled := fContReadStart		// restarts timer ticks, if continuous read
            this.PortStatus.CharsNotSent += cWstring.length - nCharsSent - cSubString.length
				return false
			endif

			if fContReadStart				// continuous read was started ?
				this.nativeSessionTimerLoop( )	// read port and process user ISR
			endif

         nCharsSent += cSubstring.length

		endfor
		this.SCtimer.enabled := fContReadStart		// restarts timer ticks, if continuous read
		return true


	//----------------------------- sendBreak -----------------------------------
	// Sends a communication BREAK on the line. Default duration will be 
	// one second, when not specified.
	function sendBreak( duration )
		local nd
		if not this.IsPortOpen( )
			return false
		endif
		nd = iif( argcount( ) == 0, 1, duration )
		if not this.chk( SetCommBreak( this.hSerial ) )
			return false
		endif
		sleep nd
		if not this.chk( ClearCommBreak( this.hSerial ) )
			return false
		endif
		return true

	//------------------------------- flushPort ---------------------------------
	// Flushes the communication port, read and write.
	function flushPort
		if not this.IsPortOpen( )
			return false
		endif
		if not this.chk( PurgeComm( this.hSerial, PURGE_TXABORT + ;
																PURGE_RXABORT + ;
																PURGE_TXCLEAR + ;
																PURGE_RXCLEAR ))
			return false
		endif
		return true

	//----------------------------- flushReadPort -------------------------------
	// Flushes the communication port, read line.
	function flushReadPort
		if not this.IsPortOpen( )
			return false
		endif
		// When port is flushed, readstring will return ''
		do while this.readString( BIGBUFFERSIZE ).length > 0
		enddo
		return true

	//--------------------------------- chk -------------------------------------
	// this function checks the return value of the Win 32 API calls.
	function chk( stat )
		if not stat
			if this.displayMessages
				AlertMessage( 'Error in Win 32 API return call', ERRORTITLE )
			endif
			return false
		endif
		return true

	//---------------------------- serialOnClose --------------------------------
	// this function, linked to the SCPaintBox onClose event, checks whether the 
	// user has properly closed the comm Port before closing the container or the form. 
	// If not, it asks for that.
	// Be aware that unclosing the comm port will make it unaccessible after. You
	// will need to restart either VdB or your Compiled .EXE application.
	function serialOnClose
		if ( this.serialCommRef.hSerial <> MINUS_ONE )
			this.serialCommRef.stopContinuousRead( )		// stop continuous read in any case
			if this.serialCommRef.displayMessages
				this.ret = ConfirmationMessage( 'The Comm port has not been closed. Do you want to close it now ?' ,;
								 'SerialComm Warning !' )
				if ( this.ret == BUTTON_YES )
					this.serialCommRef.closePort( )
				else
					return false
				endif
			else
				return false
			endif
		endif
		return true

	//---------------------------- IsPortOpen -----------------------------------
	// this function returns TRUE if the port is opened, false otherwise
	// It displays a message after trying to execute a function while the 
	// com port is not yet opened. The message is displayed if the displayMessages 
	// property of the component is set to true 
	function IsPortOpen
		if ( this.hSerial == MINUS_ONE )
			if this.displayMessages
				AlertMessage( 'Cannot execute function, the Com Port is not opened', ERRORTITLE )
			endif
			return false
		endif
		return true

endclass

// class for the CommErrors property
class CommErrorClass
	this.type = new array( )            // Types of error
	this.text = new array( )            // Error text strings
endclass

// class for the PortStatus property 
class PortStatusClass
	this.CTSHold = false						// Tx waiting for CTS signal (If hardware handshake)
	this.DSRHold = false						// Tx waiting for DSR signal (If hardware handshake)
	this.XoffHold = false					// Tx waiting, XOFF char rec'd
	this.XoffSent = false					// Tx waiting, XOFF char sent 
	this.CharsInRxQ = 0	  					// # of bytes in input buffer, not yet read
	this.CharsNotSent = 0					// # of bytes not sent after a write due to an error.
endclass

// class for the XonXoffSet property
class XonXoffConfigClass
	this.XonChar = DEFAULT_XON 			// Xon char currently set
	this.XoffChar = DEFAULT_XOFF        // Xoff char currently set
	this.XonPercent = 0						// Percentage of receive buffer size that trigs Xon
	this.XoffPercent = 0						// Percentage of receive buffer size that trigs Xoff
	this.InBufSize = Null					// Actual receive buffer size
endclass


/*------------------------------------------------------------------------------
     This section contains Windows API C++ type structure declarations 
  	  that are used for some Win 32 API calls involving communication ports. 
  	  These structure classes are inherited from a base class defined in the
  	  file StructureEx.cc
--------------------------------------------------------------------------------*/

#include <StructApi.h>
#define TYPE_DCB   TYPE_STRING

class _COMMTIMEOUTS of structureEx
	super::addMember( TYPE_DWORD,   "ReadIntervalTimeout" )
	super::addMember( TYPE_DWORD,   "ReadTotalTimeoutMultiplier" )
	super::addMember( TYPE_DWORD,   "ReadTotalTimeoutConstant" )
	super::addMember( TYPE_DWORD,   "WriteTotalTimeoutMultiplier" )
	super::addMember( TYPE_DWORD,   "WriteTotalTimeoutConstant" )
endclass

class _COMSTAT of structureEx
	super::addMember( TYPE_DWORD,   "fCtsHold", 1 )          // Tx waiting for CTS signal
	super::addMember( TYPE_DWORD,   "fDsrHold", 1 )          // Tx waiting for DSR signal 
	super::addMember( TYPE_DWORD,   "fRlsdHold", 1 )         // Tx waiting for RLSD signal
	super::addMember( TYPE_DWORD,   "fXoffHold", 1 )         // Tx waiting, XOFF char rec'd
	super::addMember( TYPE_DWORD,   "fXoffSent", 1 )         // Tx waiting, XOFF char sent
	super::addMember( TYPE_DWORD,   "fEof", 1 )              // EOF character sent
	super::addMember( TYPE_DWORD,   "fTxim", 1 )             // character waiting for Tx
	super::addMember( TYPE_DWORD,   "fReserved", 25 )        // reserved
	super::addMember( TYPE_DWORD,   "cbInQue" )              // bytes in input buffer
	super::addMember( TYPE_DWORD,   "cbOutQue" )             // bytes in output buffer
endclass

class _DCB of structureEx
	super::addMember( TYPE_DWORD,   "DCBlength" )            // sizeof(DCB) 
	super::addMember( TYPE_DWORD,   "BaudRate" )             // current baud rate
	super::addMember( TYPE_DWORD,   "fBinary", 1 )           // binary mode, no EOF check
	super::addMember( TYPE_DWORD,   "fParity", 1 )           // enable parity checking
	super::addMember( TYPE_DWORD,   "fOutxCtsFlow", 1 )      // CTS output flow control
	super::addMember( TYPE_DWORD,   "fOutxDsrFlow", 1 )      // DSR output flow control
	super::addMember( TYPE_DWORD,   "fDtrControl", 2 )       // DTR flow control type
	super::addMember( TYPE_DWORD,   "fDsrSensitivity", 1 )   // DSR sensitivity 
	super::addMember( TYPE_DWORD,   "fTXContinueOnXoff", 1 ) // XOFF continues Tx
	super::addMember( TYPE_DWORD,   "fOutX", 1 )             // XON/XOFF out flow control
	super::addMember( TYPE_DWORD,   "fInX", 1 )              // XON/XOFF in flow control
	super::addMember( TYPE_DWORD,   "fErrorChar", 1 )        // enable error replacement
	super::addMember( TYPE_DWORD,   "fNull", 1 )             // enable null stripping
	super::addMember( TYPE_DWORD,   "fRtsControl", 2 )       // RTS flow control
	super::addMember( TYPE_DWORD,   "fAbortOnError", 1 )     // abort reads/writes on error 
	super::addMember( TYPE_DWORD,   "fDummy2", 17 )          // reserved 
	super::addMember( TYPE_WORD,    "wReserved" )            // not currently used
	super::addMember( TYPE_WORD,    "XonLim" )               // transmit XON threshold
	super::addMember( TYPE_WORD,    "XoffLim" )              // transmit XOFF threshold
	super::addMember( TYPE_BYTE,    "ByteSize" )             // number of bits/byte, 4-8 
	super::addMember( TYPE_BYTE,    "Parity" )               // 0-4=no,odd,even,mark,space
	super::addMember( TYPE_BYTE,    "StopBits" )             // 0,1,2 = 1, 1.5, 2 
	super::addMember( TYPE_CHAR,    "XonChar" )              // Tx and Rx XON character
	super::addMember( TYPE_CHAR,    "XoffChar" )             // Tx and Rx XOFF character
	super::addMember( TYPE_CHAR,    "ErrorChar" )            // error replacement character
	super::addMember( TYPE_CHAR,    "EofChar" )              // end of input character 
	super::addMember( TYPE_CHAR,    "EvtChar" )              // received event character 
	super::addMember( TYPE_WORD,    "wReserved1" )           // reserved; do not use
endclass

class _COMMCONFIG of structureEx
	super::addMember( TYPE_DWORD,   "dwSize" )
	super::addMember( TYPE_WORD,    "wVersion" )
	super::addMember( TYPE_WORD,    "wReserved" )
	super::addMember( TYPE_DCB,     "dcb", new _DCB().length() )      // sizeOf( DCB )
	super::addMember( TYPE_DWORD,   "dwProviderSubType" )
	super::addMember( TYPE_DWORD,   "dwProviderOffset" )
	super::addMember( TYPE_DWORD,   "dwProviderSize" )
	super::addMember( TYPE_STRING,    "wcProviderData[1]", 128 )
endclass

class _COMMPROP of structureEx
	super::addMember( TYPE_WORD,    "wPacketLength" )       	// packet size, in bytes 
	super::addMember( TYPE_WORD,    "wPacketVersion" )      	// packet version 
	super::addMember( TYPE_DWORD,   "dwServiceMask" )      	// services implemented
	super::addMember( TYPE_DWORD,   "dwReserved1" )        	// reserved 
	super::addMember( TYPE_DWORD,   "dwMaxTxQueue" )       	// max Tx bufsize, in bytes
	super::addMember( TYPE_DWORD,   "dwMaxRxQueue" )       	// max Rx bufsize, in bytes 
	super::addMember( TYPE_DWORD,   "dwMaxBaud" )          	// max baud rate, in bps 
	super::addMember( TYPE_DWORD,   "dwProvSubType" )      	// specific provider type
	super::addMember( TYPE_DWORD,   "dwProvCapabilities" ) 	// capabilities supported 
	super::addMember( TYPE_DWORD,   "dwSettableParams" )   	// changable parameters 
	super::addMember( TYPE_DWORD,   "dwSettableBaud" )     	// allowable baud rates
	super::addMember( TYPE_WORD,    "wSettableData" )      	// allowable byte sizes
	super::addMember( TYPE_WORD,    "wSettableStopParity" ) 	// stop bits/parity allowed
	super::addMember( TYPE_DWORD,   "dwCurrentTxQueue" )    	// Tx buffer size, in bytes
	super::addMember( TYPE_DWORD,   "dwCurrentRxQueue" )     // Rx buffer size, in bytes 
	super::addMember( TYPE_DWORD,   "dwProvSpec1" )          // provider-specific data
	super::addMember( TYPE_DWORD,   "dwProvSpec2" )          // provider-specific data
	super::addMember( TYPE_WORD,    "wcProvChar[1]" )        // provider-specific data 
endclass

class _MODEMDEVCAPS of structureEx
	super::addMember( TYPE_DWORD,		"dwActualSize" )               // size of returned data, in bytes
	super::addMember( TYPE_DWORD,		"dwRequiredSize" )             // total size of structure
	super::addMember( TYPE_DWORD,		"dwDevSpecificOffset" )        // offset of provider-defined data
	super::addMember( TYPE_DWORD,		"dwDevSpecificSize" )          // size of provider-defined data

	 // Product and version identification
	super::addMember( TYPE_DWORD,		"dwModemProviderVersion" )     // provider version number
	super::addMember( TYPE_DWORD,		"dwModemManufacturerOffset" )  // offset of manufacturer name
	super::addMember( TYPE_DWORD,		"dwModemManufacturerSize" )    // length of manufacturer name
	super::addMember( TYPE_DWORD,		"dwModemModelOffset" )         // offset of model name
	super::addMember( TYPE_DWORD,		"dwModemModelSize" )           // length of model name
	super::addMember( TYPE_DWORD,		"dwModemVersionOffset" )       // offset of version name
	super::addMember( TYPE_DWORD,		"dwModemVersionSize" )         // length of version name

	 // Local option capabilities
	super::addMember( TYPE_DWORD,		"dwDialOptions" )              // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwCallSetupFailTimer" )       // maximum in seconds
	super::addMember( TYPE_DWORD,		"dwInactivityTimeout" )        // maximum in tenths of seconds
	super::addMember( TYPE_DWORD,		"dwSpeakerVolume" )            // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwSpeakerMode" )              // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwModemOptions" )             // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwMaxDTERate" )               // maximum value in bit/s
	super::addMember( TYPE_DWORD,		"dwMaxDCERate" )               // maximum value in bit/s

	 // Variable portion for strings and provider-specific data
	super::addMember( TYPE_BYTE,		"abVariablePortion[1]" )       // variable-length data
endclass

class _MODEMSETTINGS of structureEx
	super::addMember( TYPE_DWORD,		"dwActualSize" )               // size of returned data, in bytes
	super::addMember( TYPE_DWORD,		"dwRequiredSize" )             // total size of structure
	super::addMember( TYPE_DWORD,		"dwDevSpecificOffset" )        // offset of provider-defined data
	super::addMember( TYPE_DWORD,		"dwDevSpecificSize" )          // size of provider-defined data

	 // Static local options (read/write)
	super::addMember( TYPE_DWORD,		"dwCallSetupFailTimer" )       // call setup timeout, in seconds
	super::addMember( TYPE_DWORD,		"dwInactivityTimeout" )        // inactivity timeout, in tenths of seconds
	super::addMember( TYPE_DWORD,		"dwSpeakerVolume" )            // speaker volume level
	super::addMember( TYPE_DWORD,		"dwSpeakerMode" )              // speaker mode
	super::addMember( TYPE_DWORD,		"dwPreferredModemOptions" )    // bitmap specifying preferred options

	 // negotiated options (read only) for current or last call
	super::addMember( TYPE_DWORD,		"dwNegotiatedModemOptions" )   // bitmap specifying actual options
	super::addMember( TYPE_DWORD,		"dwNegotiatedDCERate" )        // DCE rate, in bits per second

	 // Variable portion for proprietary expansion
	super::addMember( TYPE_BYTE,	   "abVariablePortion[1]" )       // variable-length data
endclass


